"""Simple Python PAM

Note that for now, there is no nicer API for the C module. That doesn't
really limit you from doing anything, but you just wont have that nice an
API.
"""

from _spypam import *
